/*-----------------------------------------------------------------------------
 * $RCSfile: ShoppingClMsgMgr.c,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include "ShoppingClMsgMgr.h"

#include <OpenSOAP/Transport.h>

#include <string.h>

#define READ_BUF_SZ 256

#define APPLICATION_ERROR (0x10000001L)

/* error message macro */
#define ERROR_MSG(error, location, message); \
fprintf(stderr, \
		"%s function error: Cannot %s\n" \
		" ---> OpenSOAP Error Code: 0x%04x\n", \
		(location),\
		(message),\
		(error));

/* error check macro */
#define ERROR_CHECK(error, location, message); \
if (OPENSOAP_FAILED(error)) { \
	ERROR_MSG(error, location, message) \
}

/* error return macro */
#define ERROR_RETURN(error, location, message); \
if (OPENSOAP_FAILED(error)) { \
	ERROR_MSG(error, location, message) \
	return (error); \
}

static
const unsigned char
NULL_CHAR = '\0';

static const char OPENSOAP_HEADER_NS_URI[] = "http://header.opensoap.jp/1.0/";
static const char OPENSOAP_HEADER_NS_PREFIX[] = "opensoap-header";

static const char METHOD_NS_URI[]
="http://services.opensoap.jp/samples/Shopping/";
static const char METHOD_NS_PREFIX[] = "m";

static int ttl = 80;

/* read message from text file as byte array */
static
int
ReadMessage(FILE *fp,
			const char *charEnc,
			OpenSOAPEnvelopePtr *env) {
	static char FuncName[] = "ReadMessage";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPByteArrayPtr ba;
	
	error = (fp ? error : APPLICATION_ERROR);
	ERROR_RETURN(error, FuncName, "find input file pointer");
	
	error = OpenSOAPByteArrayCreate(&ba);
	if (OPENSOAP_SUCCEEDED(error)) {
		unsigned char readBuf[READ_BUF_SZ];
		size_t readSize = 0;
		int relError = OPENSOAP_NO_ERROR;
		
		while (OPENSOAP_SUCCEEDED(error)
			   && (readSize = fread(readBuf, 1, READ_BUF_SZ, fp))) {
			error = OpenSOAPByteArrayAppend(ba, readBuf, readSize);
		}
		
		if (OPENSOAP_SUCCEEDED(error)) {
			
			error = OpenSOAPEnvelopeCreateCharEncoding(charEnc, ba, env);
			ERROR_CHECK(error, FuncName, "create envelope from byte array");
			
		}
		else {
			ERROR_MSG(error, FuncName, "complete reading text as byte array");
		}
		
		relError = OpenSOAPByteArrayRelease(ba);
		ERROR_RETURN(relError, FuncName, "release byte array");
	}
	else {
		ERROR_MSG(error, FuncName, "create byte array");
	}
	
	return error;
}

/* write message envelope as text file */
static
int
WriteMessage(OpenSOAPEnvelopePtr env,
			 const char *charEnc,
			 FILE *fp) {
	static char FuncName[] = "WriteMessage";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPByteArrayPtr envBuf = NULL;
	
	error = (env ? error : APPLICATION_ERROR);
	ERROR_RETURN(error, FuncName, "find envelope");
	
	error = (fp ? error : APPLICATION_ERROR);
	ERROR_RETURN(error, FuncName, "find output file pointer");
	
	error = OpenSOAPByteArrayCreate(&envBuf);
	if (OPENSOAP_SUCCEEDED(error)) {
		
		error = OpenSOAPEnvelopeGetCharEncodingString(env, charEnc, envBuf);
		if (OPENSOAP_SUCCEEDED(error)) {
			const unsigned char *envBeg = NULL;
			size_t envSz = 0;
			
			error = OpenSOAPByteArrayGetBeginSizeConst(envBuf,
													   &envBeg,
													   &envSz);
			if (OPENSOAP_SUCCEEDED(error)) {
				fwrite(envBeg, 1, envSz, fp);
			}
			else {
				ERROR_MSG(error, FuncName, "get begin&size of byte array");
			}
		}
		else {
			ERROR_MSG(error, FuncName, "get character encoding string");
		}
		
		error = OpenSOAPByteArrayRelease(envBuf);
		ERROR_RETURN(error, FuncName, "release byte array");
	}
	else {
		ERROR_MSG(error, FuncName, "create byte array");
	}
	
	return error;
}

/* write message envelope with label string as text file */
WriteLabeledMessage(OpenSOAPEnvelopePtr env,
					const char *charEnc,
					const char *label,
					FILE *fp) {
	static char FuncName[] = "WriteLabeledMessage";
    int error = OPENSOAP_NO_ERROR;
	
	if (!label) {
		label = "";
	}
	
	fprintf(fp, "\n===== begin %s soap message =====\n", label);
	
	error = WriteMessage(env, charEnc, fp);
	ERROR_RETURN(error, FuncName, "release byte array");
	
	fprintf(fp, "\n===== end %s soap message =====\n", label);
	
	return error;
}

/* soap invoke function */
static
int
InvokeService(const char *endpoint,
			  const char *soapAction,
			  const char *charEnc,
			  OpenSOAPEnvelopePtr *request,
			  OpenSOAPEnvelopePtr *response) {
	static char FuncName[] = "InvokeService";
    int error = OPENSOAP_NO_ERROR;
    OpenSOAPTransportPtr transport = NULL;
	
    error = OpenSOAPTransportCreate(&transport);
	if (OPENSOAP_SUCCEEDED(error)) {
	
		error = OpenSOAPTransportSetURL(transport, endpoint);
		if (OPENSOAP_SUCCEEDED(error)) {
			
			error = OpenSOAPTransportSetSOAPAction(transport, soapAction);
			if (OPENSOAP_SUCCEEDED(error)) {
				
				error = OpenSOAPTransportSetCharset(transport, charEnc);
				if (OPENSOAP_SUCCEEDED(error)) {
					
					error = WriteLabeledMessage(*request,
												charEnc,
												"request",
												stderr);
					if (OPENSOAP_SUCCEEDED(error)) {
						
						error = OpenSOAPTransportInvoke(transport,
														*request,
														response);
						if (OPENSOAP_SUCCEEDED(error)) {
							
							error = WriteLabeledMessage(*response,
														NULL,
														"response",
														stderr);
							ERROR_CHECK(error, FuncName,
										"output response message");
						}
						else {
							ERROR_MSG(error, FuncName, "invoke transport");
						}
					}
					else {
						ERROR_MSG(error, FuncName, "output request message");
					}
				}
				else {
					ERROR_MSG(error, FuncName, "set Charset");
				}
			}
			else {
				ERROR_MSG(error, FuncName, "set SOAPAction");
			}
		}
		else {
			ERROR_MSG(error, FuncName, "set URL");
		}
		
		error = OpenSOAPTransportRelease(transport);
		ERROR_RETURN(error, FuncName, "release transport");
	}
	else {
		ERROR_MSG(error, FuncName, "create transport");
	}		

    return error;
}

/* add attribute to element */
static
int
AddAttribute(OpenSOAPBlockPtr block,
			 const char *elmName,
			 const char *attrName,
			 const char *attrType,
			 void *attrValue,
			 OpenSOAPXMLAttrPtr *attr) {
	static char FuncName[] = "AddAttribute";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPXMLElmPtr elm = NULL;
	
    error = OpenSOAPBlockGetChildMB(block, elmName, &elm);
	ERROR_RETURN(error, FuncName, "get specifed element");
	
    error = OpenSOAPXMLElmAddAttributeMB(elm,
										 attrName,
										 attrType,
										 attrValue,
										 attr);
	ERROR_RETURN(error, FuncName, "add attribute to element");
	
	return error;
}

/* add attribute with namespace to element */
static
int
AddAttributeWithNamespace(OpenSOAPBlockPtr block,
						  const char *elmName,
						  const char *attrName,
						  const char *attrType,
						  void *attrValue,
						  const char *NsUri,
						  const char *NsPrefix) {
	static char FuncName[] = "AddAttributeWithNamespace";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPXMLAttrPtr attr = NULL;
	
	error = AddAttribute(block, elmName, attrName, attrType, attrValue, &attr);
	ERROR_RETURN(error, FuncName, "add attribute to element");
	
    error = OpenSOAPXMLAttrSetNamespaceMB(attr, NsUri, NsPrefix);
	ERROR_RETURN(error, FuncName, "set namespace to attribute");
	
	return error;
}

/* create common part of request message */
static
int
CreateRequestCommon(OpenSOAPEnvelopePtr *request,
					const char *requestName,
					const char *async,
					OpenSOAPBlockPtr *body) {
	static char FuncName[] = "CreateRequestCommon";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPBlockPtr header;
	
    error = OpenSOAPEnvelopeCreateMB("1.1", NULL, request);
	ERROR_RETURN(error, FuncName, "create envelope");
	
	if (strcmp(async, "true") == 0) {
		
		OpenSOAPXMLElmPtr ttlElm = NULL;
		OpenSOAPStringPtr typeStr = NULL;
		OpenSOAPStringPtr asyncStr = NULL;
		OpenSOAPXMLElmPtr asyncElm = NULL;
		
		/* add opensoap-header block */
		error = OpenSOAPEnvelopeAddHeaderBlockMB(*request,
												 "opensoap-header-block",
												 &header); 
		ERROR_RETURN(error, FuncName, "add header part");
		
		/* set namespace to opensoap-header block */
		error = OpenSOAPBlockSetNamespaceMB(header,
											OPENSOAP_HEADER_NS_URI,
											OPENSOAP_HEADER_NS_PREFIX);
		ERROR_RETURN(error, FuncName, "set namespace to header part");
		
		/* set ttl element to opensoap-header */
		error = OpenSOAPBlockSetChildValueMB(header, "ttl", "int", &ttl);
		ERROR_RETURN(error, FuncName, "set ttl element");
		
		/* set namespace to ttl element */
		error = OpenSOAPBlockGetChildMB(header, "ttl", &ttlElm);
		ERROR_RETURN(error, FuncName, "get ttl element");
		
		error = OpenSOAPXMLElmSetNamespaceMB(ttlElm,
											 OPENSOAP_HEADER_NS_URI,
											 OPENSOAP_HEADER_NS_PREFIX);
		ERROR_RETURN(error, FuncName, "set namespace to ttl element");
		
		/* add type attribute with namespace to ttl element */
		
		error = OpenSOAPStringCreateWithMB("second", &typeStr);
		ERROR_RETURN(error, FuncName, "create string for type attr");
		
		error = AddAttributeWithNamespace(header,
										  "ttl",
										  "type",
										  "string",
										  &typeStr,
										  OPENSOAP_HEADER_NS_URI,
										  OPENSOAP_HEADER_NS_PREFIX);
 		ERROR_RETURN(error, FuncName, "add type attr to ttl element");
		
		/* add async element */
		
		error = OpenSOAPStringCreateWithMB(async, &asyncStr);
		ERROR_RETURN(error, FuncName, "create string for async");
		
		error = OpenSOAPBlockSetChildValueMB(header,
											 "async",
											 "string",
											 &asyncStr);
		ERROR_RETURN(error, FuncName, "set async element");
		
		/* set namespace to async element */
		
		error = OpenSOAPBlockGetChildMB(header, "async", &asyncElm);
		ERROR_RETURN(error, FuncName, "get async element");
		
		error = OpenSOAPXMLElmSetNamespaceMB(asyncElm,
											 OPENSOAP_HEADER_NS_URI,
											 OPENSOAP_HEADER_NS_PREFIX);
		ERROR_RETURN(error, FuncName, "set namespace to async element");
	}
	
    /* add method body block */
    error = OpenSOAPEnvelopeAddBodyBlockMB(*request, requestName, body); 
	ERROR_RETURN(error, FuncName, "add body part");
	
    /* set namespace to method body block */
    error = OpenSOAPBlockSetNamespaceMB(*body,
										METHOD_NS_URI,
										METHOD_NS_PREFIX);
	ERROR_RETURN(error, FuncName, "set namespace to body part");
	
	return error;
}

/* set shopping service method parameter: index */
static
int
SetParameterIndex(long index,
				  OpenSOAPBlockPtr *body) {
	static char FuncName[] = "SetParameterIndex";
    int error = OPENSOAP_NO_ERROR;
	
	error = OpenSOAPBlockSetChildValueMB(*body, "index", "int", &index);
	ERROR_RETURN(error, FuncName, "set parameter: index");
	
	return error;
}

/* set shopping service method parameter: code */
static
int
SetParameterCode(const char *code,
				 OpenSOAPBlockPtr *body) {
	static char FuncName[] = "SetParameterCode";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPStringPtr codeStr = NULL;
	
	error = OpenSOAPStringCreateWithMB(code, &codeStr);
	ERROR_RETURN(error, FuncName, "create string for code");
	
	error = OpenSOAPBlockSetChildValueMB(*body, "code", "string", &codeStr);
	ERROR_RETURN(error, FuncName, "set parameter: code");
	
	return error;
}

/* set shopping service method parameter: qty */
static
int
SetParameterQty(long qty,
				OpenSOAPBlockPtr *body) {
	static char FuncName[] = "SetParameterQty";
    int error = OPENSOAP_NO_ERROR;
	
	error = OpenSOAPBlockSetChildValueMB(*body, "qty", "int", &qty);
	ERROR_RETURN(error, FuncName, "set parameter: qty");
	
	return error;
}

/* get parameter value of string type from message */
static
int
GetStringParameter(OpenSOAPBlockPtr body,
				   const char *paramName,
				   char **paramValue) {
	static char FuncName[] = "GetStringParameter";
	int error = OPENSOAP_NO_ERROR;
	OpenSOAPStringPtr param = NULL;
    size_t paramLen = 0;
	
	error = OpenSOAPBlockGetChildValueMB(body, paramName, "string", &param);
	ERROR_RETURN(error, FuncName, "get value of parameter");
	
	error = OpenSOAPStringGetLengthMB(param, &paramLen);
	ERROR_RETURN(error, FuncName, "get string length of parameter");
	
	*paramValue = malloc(paramLen + 1);
	error = (*paramValue ? error : APPLICATION_ERROR);
	ERROR_RETURN(error, FuncName, "allocate memory");
	
	error = OpenSOAPStringGetStringMB(param, &paramLen, *paramValue);
	ERROR_RETURN(error, FuncName, "get string from parameter");
	
	return error;
}

/* parse message in case of fault*/
static
int
ParseFaultMessage(OpenSOAPBlockPtr body) {
	static char FuncName[] = "ParseFaultMessage";
	int error = OPENSOAP_NO_ERROR;
	char *paramValue = NULL;
	
	fprintf(stderr, "Fault message:\n");
	
	/* parse faultcode */
	
	error = GetStringParameter(body, "faultcode", &paramValue);
	ERROR_RETURN(error, FuncName, "get faultcode");
	
	fprintf(stderr, "    faultcode: %s\n", paramValue);
	
	free(paramValue);
	paramValue = NULL;
	
	/* parse faultstring */
	
	error = GetStringParameter(body, "faultstring", &paramValue);
	ERROR_RETURN(error, FuncName, "get faultstring");
	
	fprintf(stderr, "    faultstring: %s\n", paramValue);
	
	free(paramValue);
	paramValue = NULL;
	
	/* parse faultactor */
	
	error = GetStringParameter(body, "faultactor", &paramValue);
	ERROR_RETURN(error, FuncName, "get faultactor");
	
	fprintf(stderr,
			"    faultactor: %s\n",
			paramValue ? paramValue : "(null)");
	
	free(paramValue);
	paramValue = NULL;
	
	/* parse detail (as plane string) */
	
	error = GetStringParameter(body, "detail", &paramValue);
	ERROR_RETURN(error, FuncName, "get detail");
	
	fprintf(stderr, "    detail: %s\n", paramValue ? paramValue : "(null)");
	
	free(paramValue);
	paramValue = NULL;
	
	return error;
}

/* parse common part of response message */
static
int
ParseResponseCommon(OpenSOAPEnvelopePtr response,
					const char *responseName,
					OpenSOAPBlockPtr *body,
					int *isFault) {
	static char FuncName[] = "ParseResponseCommon";
    int error = OPENSOAP_NO_ERROR;
	int isSameNs = 0;
	
	*isFault = 0;
	
	/* for fault message */
	
	error = OpenSOAPEnvelopeGetBodyBlockMB(response, "Fault", body);
	ERROR_RETURN(error, FuncName, "complete getting body block");
	
	if (*body != NULL) {
		
		error = ParseFaultMessage(*body);
		ERROR_RETURN(error, FuncName, "parse fault message");
		
		*isFault = 1;
		
		return error;
	}
	
	/* process response body */
	
	error = OpenSOAPEnvelopeGetBodyBlockMB(response, responseName, body);
	ERROR_RETURN(error, FuncName, "complete getting body block");
	
	if (*body != NULL) {
		
		error = OpenSOAPBlockIsSameNamespaceMB(*body,
											   METHOD_NS_URI,
											   &isSameNs);
		ERROR_RETURN(error, FuncName, "complete namespace matching");
		
		error = (isSameNs ? error : APPLICATION_ERROR);
		ERROR_RETURN(error, FuncName, "match namespace");
		
		return error;
	}
	
	return APPLICATION_ERROR;
}

/* create get product count request message */
static
int
CreateGetProductCountRequest(OpenSOAPEnvelopePtr *request) {
	static char FuncName[] = "CreateGetProductCountRequest";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPBlockPtr body = NULL;
	
	error = CreateRequestCommon(request, "GetProductCount", "false", &body);
	ERROR_RETURN(error, FuncName, "create request");
	
	return error;
}

/* parse get product count response message */
static
int
ParseGetProductCountResponse(OpenSOAPEnvelopePtr response,
							 long *count) {
	static char FuncName[] = "ParseGetProductCountResponse";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPBlockPtr body = NULL;
	int isFault = 0;
	
	error = ParseResponseCommon(response,
								"GetProductCountResponse",
								&body,
								&isFault);
	ERROR_RETURN(error, FuncName, "parse common part of response");
	
	if (isFault) {
		return error;
	}
	
	error = OpenSOAPBlockGetChildValueMB(body, "count", "int", count);
	ERROR_RETURN(error, FuncName, "get parameter: count");
	
	return error;
}

/* submit get product count request */
extern
int
GetProductCount(const char *endpoint,
				const char *soapAction,
				const char *charEnc,
				long *count) {
	static char FuncName[] = "GetProductCount";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPEnvelopePtr request = NULL;

	*count = -1;
	
	error = CreateGetProductCountRequest(&request);
	if (OPENSOAP_SUCCEEDED(error)) {
		OpenSOAPEnvelopePtr response = NULL;
		
		error = InvokeService(endpoint,
							  soapAction,
							  charEnc,
							  &request,
							  &response);
		if (OPENSOAP_SUCCEEDED(error)) {
			
			error = ParseGetProductCountResponse(response, count);
			ERROR_CHECK(error, FuncName, "parse response message");
			
			error = OpenSOAPEnvelopeRelease(response);
			ERROR_CHECK(error, FuncName, "release response envelope");
		} else {
			ERROR_MSG(error, FuncName, "invoke service");
		}			
		error = OpenSOAPEnvelopeRelease(request);
		ERROR_RETURN(error, FuncName, "release request envelope");
	}
	else {
 		ERROR_MSG(error, FuncName, "create request message");
	}
	
	return error;
}

/* create get product specification request message */
static
int
CreateGetProductSpecRequest(long index,
							OpenSOAPEnvelopePtr *request) {
	static char FuncName[] = "CreateGetProductSpecRequest";
    int error = OPENSOAP_NO_ERROR;
    OpenSOAPBlockPtr body = NULL;
	
	error = CreateRequestCommon(request, "GetProductSpec", "false", &body);
	ERROR_RETURN(error, FuncName, "create common part of request");
	
	error = SetParameterIndex(index, &body);
	ERROR_RETURN(error, FuncName, "set parameter: index");
	
	return error;
}

/* parse get product specification response message */
static
int
ParseGetProductSpecResponse(OpenSOAPEnvelopePtr response,
							ProductListItemPtr productList) {
	static char FuncName[] = "ParseGetProductSpecResponse";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPBlockPtr body = NULL;
	int isFault = 0;
	char *manufacturer = NULL;
	char *name = NULL;
	char *code = NULL;
	long price = -1;
	
	error = ParseResponseCommon(response,
								"GetProductSpecResponse",
								&body,
								&isFault);
	ERROR_RETURN(error, FuncName, "parse common part of response");
	
	if (isFault) {
		return error;
	}
	
	error = GetStringParameter(body, "manufacturer", &manufacturer);
	ERROR_RETURN(error, FuncName, "get parameter: manufacturer");
	
	error = GetStringParameter(body, "name", &name);
	ERROR_RETURN(error, FuncName, "get parameter: name");
	
	error = GetStringParameter(body, "code", &code);
	ERROR_RETURN(error, FuncName, "get parameter: code");
	
	error = OpenSOAPBlockGetChildValueMB(body, "price", "int", &price);
	ERROR_RETURN(error, FuncName, "get parameter: price");
	
	fprintf(stderr,
			"\nGetProductSpecResponse:\n"
			"    manufacturer=%s\n"
			"    name=%s\n"
			"    code=%s\n"
			"    price=%ld\n\n",
			manufacturer ? manufacturer : "(null)",
			name ? name : "(null)",
			code ? code : "(null)",
			price);
	
	if (!manufacturer || !name || !code || price == -1) {
		error = APPLICATION_ERROR;
		ERROR_MSG(error, FuncName, "get all of parameters");
	}
	
	if (!ProductListItemSetValues(productList,
								  manufacturer,
								  name,
								  code,
								  price)) {
		error = APPLICATION_ERROR;
		ERROR_MSG(error, FuncName, "alloc memory");
	}
	
	free(code);
	free(name);
	free(manufacturer);
	
	return error;
}

/* submit get product specification request */
extern
int
GetProductSpec(const char *endpoint,
			   const char *soapAction,
			   const char *charEnc,
			   long index,
			   ProductListItemPtr productList) {
	static char FuncName[] = "GetProductSpec";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPEnvelopePtr request = NULL;
	
	error = CreateGetProductSpecRequest(index, &request);
	if (OPENSOAP_SUCCEEDED(error)) {
		OpenSOAPEnvelopePtr response = NULL;
		
		error = InvokeService(endpoint,
							  soapAction,
							  charEnc,
							  &request,
							  &response);
		if (OPENSOAP_SUCCEEDED(error)) {
			
			error = ParseGetProductSpecResponse(response, productList);
			ERROR_CHECK(error, FuncName, "parse response message");
			
			error = OpenSOAPEnvelopeRelease(response);
			ERROR_CHECK(error, FuncName, "release response envelope");
		} else {
			ERROR_MSG(error, FuncName, "invoke service");
		}
		error = OpenSOAPEnvelopeRelease(request);
		ERROR_RETURN(error, FuncName, "release request envelope");
	}
	else {
		ERROR_MSG(error, FuncName, "create request message");
	}
	
    return error;
}

/* create get stock quantity request message */
static
int
CreateGetStockQtyRequest(const char *code,
						 OpenSOAPEnvelopePtr *request) {
	static char FuncName[] = "CreateGetStockQtyRequest";
    int error = OPENSOAP_NO_ERROR;
    OpenSOAPBlockPtr body = NULL;
	
	error = CreateRequestCommon(request, "GetStockQty", "false", &body);
	ERROR_RETURN(error, FuncName, "create common part of request");
	
	error = SetParameterCode(code, &body);
	ERROR_RETURN(error, FuncName, "set parameter: code");
	
	return error;
}

/* parse get stock quantity response message */
static
int
ParseGetStockQtyResponse(OpenSOAPEnvelopePtr response,
						 long *qty) {
	static char FuncName[] = "ParseGetStockQtyResponse";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPBlockPtr body = NULL;
	int isFault = 0;
	
	error = ParseResponseCommon(response,
								"GetStockQtyResponse",
								&body,
								&isFault);
	ERROR_RETURN(error, FuncName, "parse common part of response");
	
	if (isFault) {
		return error;
	}
	
	error = OpenSOAPBlockGetChildValueMB(body, "qty", "int", qty);
	ERROR_RETURN(error, FuncName, "get parameter: qty");
	
	return error;
}

/* submit get stock quantity request */
extern
int
GetStockQty(const char *endpoint,
			const char *soapAction,
			const char *charEnc,
			const char *code,
			long *qty) {
	static char FuncName[] = "GetStockQty";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPEnvelopePtr request = NULL;
	
	*qty = -1;
	
	error = CreateGetStockQtyRequest(code, &request);
	if (OPENSOAP_SUCCEEDED(error)) {
		OpenSOAPEnvelopePtr response = NULL;
		
		error = InvokeService(endpoint,
							  soapAction,
							  charEnc,
							  &request,
							  &response);
		if (OPENSOAP_SUCCEEDED(error)) {
			
			error = ParseGetStockQtyResponse(response, qty);
			ERROR_CHECK(error, FuncName, "parse response message");
			
			error = OpenSOAPEnvelopeRelease(response);
  			ERROR_CHECK(error, FuncName, "release response envelope");
		} else {
			ERROR_MSG(error, FuncName, "invoke service");
		}
		error = OpenSOAPEnvelopeRelease(request);
		ERROR_RETURN(error, FuncName, "release request envelope");
	}
	else {
		ERROR_MSG(error, FuncName, "create request message");
	}
	
    return error;
}

/* create place order request message */
static
int
CreatePlaceOrderRequest(const char *code,
						long qty,
						const char *async,
						OpenSOAPEnvelopePtr *request) {
	static char FuncName[] = "CreatePlaceOrderRequest";
    int error = OPENSOAP_NO_ERROR;
    OpenSOAPBlockPtr body = NULL;
	
	error = CreateRequestCommon(request, "PlaceOrder", async, &body);
	ERROR_RETURN(error, FuncName, "create common part of request");
	
	error = SetParameterCode(code, &body);
	ERROR_RETURN(error, FuncName, "set parameter: code");
	
	error = SetParameterQty(qty, &body);
	ERROR_RETURN(error, FuncName, "set parameter: qty");
	
	return error;
}

/* parse place order response message */
static
int
ParsePlaceOrderResponse(OpenSOAPEnvelopePtr response,
						char **confirmation) {
	static char FuncName[] = "ParsePlaceOrderResponse";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPBlockPtr body = NULL;
	int isFault = 0;
	
	error = ParseResponseCommon(response,
								"PlaceOrderResponse",
								&body,
								&isFault);
	ERROR_RETURN(error, FuncName, "parse common part of response");

	if (isFault) {
		return error;
	}
	
	error = GetStringParameter(body, "confirmation", confirmation);
	ERROR_RETURN(error, FuncName, "get parameter: confirmation");
	
	return error;
}

/* submit place order request */
/* return value should be free()'d after using */
extern
int
PlaceOrder(const char *endpoint,
		   const char *soapAction,
		   const char *charEnc,
		   const char *code,
		   long qty,
		   char **confirmation) {
	static char FuncName[] = "PlaceOrder";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPEnvelopePtr request = NULL;
	
	*confirmation = NULL;
	
	error = CreatePlaceOrderRequest(code, qty, "false", &request);
	if (OPENSOAP_SUCCEEDED(error)) {
		OpenSOAPEnvelopePtr response = NULL;
		
		error = InvokeService(endpoint,
							  soapAction,
							  charEnc,
							  &request,
							  &response);
		if (OPENSOAP_SUCCEEDED(error)) {
			
			error = ParsePlaceOrderResponse(response, confirmation);
			ERROR_CHECK(error, FuncName, "parse resopnse message");
			
			error = OpenSOAPEnvelopeRelease(response);

			ERROR_CHECK(error, FuncName, "release response envelope");
			
		} else {
			ERROR_MSG(error, FuncName, "invoke service");
		}
		error = OpenSOAPEnvelopeRelease(request);
 		ERROR_RETURN(error, FuncName, "release request envelope");
	}
	else {
		ERROR_MSG(error, FuncName, "create request message");
	}
	
    return error;
}

/* submit place order request in asyncronized mode */
extern
int
PlaceOrderAsync(const char *endpoint,
				const char *soapAction,
				const char *charEnc,
				const char *code,
				long qty,
				FILE *fp) {
	static char FuncName[] = "PlaceOrderAsync";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPEnvelopePtr request = NULL;
	
	error = CreatePlaceOrderRequest(code, qty, "true", &request);
	if (OPENSOAP_SUCCEEDED(error)) {
		OpenSOAPEnvelopePtr response = NULL;
		
		error = InvokeService(endpoint,
							  soapAction,
							  charEnc,
							  &request,
							  &response);
		if (OPENSOAP_SUCCEEDED(error)) {
			
			error = WriteMessage(response, NULL, fp);
			ERROR_CHECK(error, FuncName, "output resopnse message to file");
			
			error = OpenSOAPEnvelopeRelease(response);
			ERROR_CHECK(error, FuncName, "release response envelope");
		} else {
			ERROR_MSG(error, FuncName, "invoke service");
		}
		error = OpenSOAPEnvelopeRelease(request);
 		ERROR_RETURN(error, FuncName, "release request envelope");
	}
	else {
		ERROR_MSG(error, FuncName, "create request message");
	}
	
    return error;
}

/* submit confirm order request in asyncronized mode */
extern
int
ConfirmOrder(FILE *fp,
			 const char *endpoint,
			 const char *soapAction,
			 const char *charEnc,
			 char **confirmation) {
	static char FuncName[] = "ConfirmOrder";
    int error = OPENSOAP_NO_ERROR;
	OpenSOAPEnvelopePtr request = NULL;
	
	*confirmation = NULL;
	
	error = ReadMessage(fp, charEnc, &request);
	ERROR_RETURN(error, FuncName, "input request message from file");
	
	if (OPENSOAP_SUCCEEDED(error)) {
		OpenSOAPEnvelopePtr response = NULL;
		
		error = InvokeService(endpoint,
							  soapAction,
							  charEnc,
							  &request,
							  &response);
		if (OPENSOAP_SUCCEEDED(error)) {
			
			error = ParsePlaceOrderResponse(response, confirmation);
			ERROR_CHECK(error, FuncName, "parse resopnse message");
			
			error = OpenSOAPEnvelopeRelease(response);
			ERROR_CHECK(error, FuncName, "release response envelope");
		} else {
			ERROR_MSG(error, FuncName, "invoke service");
		}
		error = OpenSOAPEnvelopeRelease(request);
 		ERROR_RETURN(error, FuncName, "release request envelope");
	}
	else {
		ERROR_MSG(error, FuncName, "create request message");
	}
	
    return error;
}
